@@ -372,6 +372,9 @@ void QQuickPath::processPath()
         d->_path = createPath(QPointF(), QPointF(), d->_attributes, d->pathLength, d->_attributePoints, &d->closed);
     }
 
+    if (d->simplify)
+        d->_path = d->_path.simplified();
+
     emit changed();
 }
 
@@ -712,6 +715,32 @@ void QQuickPath::invalidateSequentialHistory() const
     d->prevBez.isValid = false;
 }
 
+/*! \qmlproperty bool QtQuick::Path::simplify
+    \since 6.6
+
+    When set to true, the path will be simplified. This implies merging all subpaths that intersect,
+    creating a path where there are no self-intersections. Consecutive parallel lines will also be
+    merged. The simplified path is intended to be used with ShapePath.OddEvenFill. Bezier curves may
+    be flattened to line segments due to numerical instability of doing bezier curve intersections.
+*/
+void QQuickPath::setSimplify(bool s)
+{
+    Q_D(QQuickPath);
+    if (d->simplify == s)
+        return;
+
+    d->simplify = s;
+    processPath();
+
+    emit simplifyChanged();
+}
+
+bool QQuickPath::simplify() const
+{
+    Q_D(const QQuickPath);
+    return d->simplify;
+}
+
 /*!
     \qmlproperty size QtQuick::Path::scale
 
@@ -2649,22 +2678,21 @@ void QQuickPathMultiline::addToPath(QPainterPath &path, const QQuickPathData &)
 */
 
 /*!
-    \qmlproperty enumeration QtQuick::PathText::font.weight
+    \qmlproperty int QtQuick::PathText::font.weight
 
     Sets the font's weight.
 
     The weight can be one of:
-    \list
-    \li Font.Thin
-    \li Font.Light
-    \li Font.ExtraLight
-    \li Font.Normal - the default
-    \li Font.Medium
-    \li Font.DemiBold
-    \li Font.Bold
-    \li Font.ExtraBold
-    \li Font.Black
-    \endlist
+
+    \value Font.Thin        100
+    \value Font.ExtraLight  200
+    \value Font.Light       300
+    \value Font.Normal      400 (default)
+    \value Font.Medium      500
+    \value Font.DemiBold    600
+    \value Font.Bold        700
+    \value Font.ExtraBold   800
+    \value Font.Black       900
 
     \qml
     PathText { text: "Hello"; font.weight: Font.DemiBold }
@@ -2728,13 +2756,12 @@ void QQuickPathMultiline::addToPath(QPainterPath &path, const QQuickPathData &)
 
     Sets the capitalization for the text.
 
-    \list
-    \li Font.MixedCase - This is the normal text rendering option where no capitalization change is applied.
-    \li Font.AllUppercase - This alters the text to be rendered in all uppercase type.
-    \li Font.AllLowercase - This alters the text to be rendered in all lowercase type.
-    \li Font.SmallCaps - This alters the text to be rendered in small-caps type.
-    \li Font.Capitalize - This alters the text to be rendered with the first character of each word as an uppercase character.
-    \endlist
+    \value Font.MixedCase       no capitalization change is applied
+    \value Font.AllUppercase    alters the text to be rendered in all uppercase type
+    \value Font.AllLowercase    alters the text to be rendered in all lowercase type
+    \value Font.SmallCaps       alters the text to be rendered in small-caps type
+    \value Font.Capitalize      alters the text to be rendered with the first character of
+                                each word as an uppercase character
 
     \qml
     PathText { text: "Hello"; font.capitalization: Font.AllLowercase }
@@ -2769,6 +2796,12 @@ void QQuickPathMultiline::addToPath(QPainterPath &path, const QQuickPathData &)
     \endqml
 */
 
+/*!
+    \qmlproperty object QtQuick::PathText::font.features
+    \since 6.6
+
+    \include qquicktext.cpp qml-font-features
+*/
 void QQuickPathText::updatePath() const
 {
     if (!_path.isEmpty())
